@@ -3,6 +3,8 @@
 //////////////////////////////////////
 //   M&L 1992 algorithm 
 //   as presented in Mrode 2005
+// since nadiv >v2.15.0 uses lower_bound algorithm for matrix lookup
+//// based on c++ <algorithm>std::lower_bound 
 extern "C"{  
 
 void ainvml(
@@ -18,7 +20,7 @@ void ainvml(
 	int *nzmaxA
 ){         
 
-  int     j, k, h, cnt, sj, dj, istart;
+  int     lb, step, it, j, k, h, cnt, sj, dj, istart;
   double  ai, alphai;
   double  *AN = new double[2*n[0]];
   double  *li = new double[n[0]];
@@ -50,7 +52,7 @@ void ainvml(
           cnt++;
         }
 
-        if((dj >= g[0]) && (dj!= n[0])){ 
+        if((dj >= g[0]) && (dj!= n[0])){
           AN[cnt] = dj;
           li[dj] += 0.5*li[j];
           cnt++;
@@ -66,7 +68,7 @@ void ainvml(
         }
         for(h=0; h<cnt; h++){   // delete duplicates
           if(AN[h]==j){
-            AN[h] -= n[0]; 
+            AN[h] -= n[0];
           }
         }
       }  // end of while
@@ -84,12 +86,18 @@ void ainvml(
       xA[pA[k]] += 1.0;       // k,k
       istart = pA[sj];
       xA[istart] += 1.0;      // sire,sire (same as dam,dam)
-      for(j=istart; j<pA[sj+1]; j++){
-        if(iA[j] == k){
-          xA[j] += -1.0;      //sire/dam,k
-          break;
+      h = istart;
+      lb = pA[sj + 1] - 1 - h;
+      while(lb > 0){
+        step = lb/2;
+        it = h + step;
+        if(iA[it] < k){
+          h=++it;
+          lb-=step+1;
         }
-      }
+        else lb = step;
+      }  // end while
+      if(iA[h] == k) xA[h] += -1.0;      //sire/dam,k
     }
     // if k doesn't have two phantom parents from same genetic group
     else {
@@ -104,42 +112,67 @@ void ainvml(
          // sire,dam
          if(sj <= dj){
             if(dj != n[0]){
-               for(j=istart; j<pA[sj+1]; j++){
-                 if(iA[j] == dj){
-                   xA[j] += alphai; 
-                   break;
-                 }
-               }
+              h = istart;
+              lb = pA[sj+1] - 1 - h;
+              while(lb > 0){
+                step = lb/2;
+                it= h + step;
+                if(iA[it] < dj){
+                  h=++it;
+                  lb-=step+1;
+                }
+                else lb = step;
+              }
+              if(iA[h] == dj) xA[h] += alphai;
             }
          }
          // sire,k
-         for(j=istart; j<pA[sj+1]; j++){
-            if(iA[j] == k){
-              xA[j] += alphai * -2.0;
-              break;
-            }
+         h = istart;
+         lb = pA[sj+1] - 1 - h;
+         while(lb > 0){
+           step = lb/2;
+           it = h + step;
+           if(iA[it] < k){
+             h=++it;
+             lb-=step+1;
+           }
+           else lb = step;
          }
+         if(iA[h] == k) xA[h] += alphai * -2.0;
       }
       if(dj != n[0]){
          istart = pA[dj];
          // dam,dam
          xA[istart] += alphai;
-         for(j=istart; j<pA[dj+1]; j++){
          // sire
          // dam,k
-            if(iA[j] == k){
-              xA[j] += alphai * -2.0;
-            }
+         h = istart;
+         lb = pA[dj+1] - 1 - h;
+         while(lb > 0){
+           step = lb/2;
+           it = h + step;
+           if(iA[it] < k){
+             h=++it;
+             lb-=step+1;
+           }
+           else lb = step;
          }
+         if(iA[h] == k) xA[h] += alphai * -2.0;
          // dam,sire
          if(dj <= sj){
             if(sj != n[0]){
-               for(j=istart; j<pA[dj+1]; j++){
-                 if(iA[j] == sj){
-                   xA[j] += alphai;
-                   break;
-                 }
-               }
+              h = istart;
+              lb = pA[dj+1] - 1 - h;
+              while(lb > 0){
+                step = lb/2;
+                it = h + step;
+                if(iA[it] < sj){
+                  h=++it;
+                  lb-=step+1;
+                }
+                else lb = step;
+              }
+              if(iA[h] == sj) xA[h] += alphai;
             }
          }
        }
@@ -157,6 +190,8 @@ void ainvml(
 ///////////////////////////////////////////////////////////////////////////
 //   basically the same M&L 1992 algorithm as 'ainvml', but
 //   with fuzzy classification of genetic groups
+// since nadiv > v2.14.3 uses lower_bound algorithm for matrix lookup
+//// based on c++ <algorithm>std::lower_bound 
 extern "C"{  
 
 void ainvfuzz(
@@ -176,7 +211,7 @@ void ainvfuzz(
 	int *pA
 ){         
 
-  int     h, i, j, k, s, d, cnt, sj, dj, mj, mk, sp, dp, fistart, aistart;
+  int     lb, step, it, h, i, j, k, s, d, cnt, sj, dj, mj, mk, sp, dp, fistart, aistart;
   double  ai, alphai, pij, pijp, pik, pikp;
   double  *AN = new double[2*n[0]];
   double  *li = new double[n[0]];
@@ -274,12 +309,19 @@ void ainvfuzz(
             pij += xF[j];
             // k,sire-group s
             aistart = pA[s];
-            for(h=aistart; h<pA[s+1]; h++){
-              if(iA[h] == k){
-                xA[h] += alphai * -2.0 * pij;
-                break;
+            h = aistart;
+            lb = pA[s+1] - 1 - h;
+            while(lb > 0){
+              step = lb/2;
+              it = h + step;
+              if(iA[it] < k){
+                h=++it;
+                lb-=step+1;
               }
+              else lb = step;
             }
+            if(iA[h] == k) xA[h] += alphai * -2.0 * pij;
+
             for(sp=0; sp<=s; sp++){
               fistart = pF[sp];
               pijp=0.0; 
@@ -288,12 +330,18 @@ void ainvfuzz(
                   pijp += xF[h];
                   // group, k's sire's-group sp
                   aistart = pA[sp];
-                  for(i=aistart; i<pA[sp+1]; i++){
-                    if(iA[i] == s){
-                      xA[i] += alphai * pij * pijp;
-                      break;
+                  i = aistart;
+                  lb = pA[sp+1] - 1 - i;
+                  while(lb > 0){
+                    step = lb/2;
+                    it = i + step;
+                    if(iA[it] < s){
+                      i=++it;
+                      lb-=step+1;
                     }
-                  }  // end i for loop
+                    else lb = step;
+                  }
+                  if(iA[i] == s) xA[i] += alphai * pij * pijp;
                   break;      			// break out of h for loop
                 }  // end if(iF[h] == sj)
               }  // end h for loop
@@ -315,12 +363,19 @@ void ainvfuzz(
             pik += xF[j];
             // k,dam-group d
             aistart = pA[d];
-            for(h=aistart; h<pA[d+1]; h++){
-              if(iA[h] == k){
-                xA[h] += alphai * -2.0 * pik;
-                break;
+            h  =aistart;
+            lb = pA[d+1] - 1 - h;
+            while(lb > 0){
+              step = lb/2;
+              it = h + step;
+              if(iA[it] < k){
+                h=++it;
+                lb-=step+1;
               }
+              else lb  =step;
             }
+            if(iA[h] == k) xA[h] += alphai * -2.0 * pik;
+
             for(dp=0; dp<=d; dp++){
               fistart = pF[dp];
               pikp=0.0; 
@@ -329,13 +384,18 @@ void ainvfuzz(
                   pikp += xF[h];
                   // group, k's dam's-group dp
                   aistart = pA[dp];
-                  for(i=aistart; i<pA[dp+1]; i++){
-                    if(iA[i] == d){
-                       xA[i] += alphai * pik * pikp;
-                      break;       		// break out of i for loop
+                  i  =aistart;
+                  lb = pA[dp+1] - 1 - i;
+                  while(lb > 0){
+                    step = lb/2;
+                    it = i + step;
+                    if(iA[it] < d){
+                      i=++it;
+                      lb-=step+1;
                     }
-                  }  // end i for loop
-
+                    else lb = step;
+                  }
+                  if(iA[i] == d) xA[i] += alphai * pik * pikp;
                   break;        		// break out of h for loop
                 }
               }  // end h for loop
@@ -364,12 +424,18 @@ void ainvfuzz(
                   // k's sire's group, k's dam's group 
                   aistart = pA[min(s, d)];   
                   if(s != d){
-                    for(i=aistart; i<pA[min(s, d)+1]; i++){
-                      if(iA[i] == max(s, d)){
-                        xA[i] += alphai * pij * pik;
-                       break;  			// break out of i for loop
+                    i = aistart;
+                    lb = pA[min(s, d)+1] - 1 - i;
+                    while(lb > 0){
+                      step = lb/2;
+                      it = i + step;
+                      if(iA[it] < max(s, d)){
+                        i=++it;
+                        lb-=step+1;
                       }
-                    }  // end i for loop
+                      else lb = step;
+                    }
+                    if(iA[i] == max(s, d)) xA[i] += alphai * pij * pik;
                   }
                   else {
                     xA[aistart] += alphai * 2.0 * pij * pik;
@@ -402,7 +468,7 @@ void ainvfuzz(
           if(iF[j] == sj){
             pij += xF[j];
             aistart = pA[s];
-            for(h=aistart; h<pA[s+1]; h++){
+            for(h=aistart; h<pA[s+1]; h++){  // NOTE: not using lower_bound algorithm
               // dam of k,sire-group s
               if(iA[h] == dj){
                 xA[h] += alphai * pij;
@@ -422,13 +488,18 @@ void ainvfuzz(
                   pijp += xF[h];
                   // group, k's sire's-group sp
                   aistart = pA[sp];
-                  for(i=aistart; i<pA[sp+1]; i++){
-                    if(iA[i] == s){
-                      xA[i] += alphai * pij * pijp;
-                      break;  			// break out of i for loop
+                  i = aistart;
+                  lb = pA[sp+1] - 1 - i;
+                  while(lb > 0){
+                    step = lb/2;
+                    it = i + step;
+                    if(iA[it] < s){
+                      i=++it;
+                      lb-=step+1;
                     }
-                  }  // end i for loop
-
+                    else lb = step;
+                  }
+                  if(iA[i] == s) xA[i] += alphai * pij * pijp;
                   break;  			// break out of h for loop
                 }
               }  // end h for loop
@@ -444,12 +515,18 @@ void ainvfuzz(
       ////// the dam contributions to k
       aistart = pA[dj];
       xA[aistart] += alphai;
-      for(j=aistart; j<pA[dj+1]; j++){
-        if(iA[j] == k){
-          xA[j] += alphai * -2.0;
-         break;
+      j = aistart;
+      lb = pA[dj+1] - 1 - j;
+      while(lb > 0){
+        step = lb/2;
+        it = j + step;
+        if(iA[it] < k){
+          j=++it;
+          lb-=step+1;
         }
-      }  // end j for loop
+        else lb = step;
+      }
+      if(iA[j] == k) xA[j] += alphai * -2.0;
     }  // end if phantom sire and known dam
 
 
@@ -464,13 +541,18 @@ void ainvfuzz(
       ////// the sire contributions to k
       aistart = pA[sj];
       xA[aistart] += alphai;
-      for(j=aistart; j<pA[sj+1]; j++){
-        if(iA[j] == k){
-          xA[j] += alphai * -2.0;
-          break;
+      j = aistart;
+      lb = pA[sj+1] - 1 - j;
+      while(lb > 0){
+        step = lb/2;
+        it = j + step;
+        if(iA[it] < k){
+          j=++it;
+          lb-=step+1;
         }
-      }  // end j for loop
-
+        else lb = step;
+      }
+      if(iA[j] == k) xA[j] += alphai * -2.0;
       //////////////////////////////////
       ////// the dam contributions to k
       for(d=0; d<g[0]; d++){                   // for each dam genetic group
@@ -480,7 +562,7 @@ void ainvfuzz(
           if(iF[j] == dj){
             pik += xF[j];
             aistart = pA[d];
-            for(h=aistart; h<pA[d+1]; h++){
+            for(h=aistart; h<pA[d+1]; h++){  // NOTE: not using lower_bound algorithm
               // sire of k,dam-group d
               if(iA[h] == sj){
                 xA[h] += alphai * pik;
@@ -499,13 +581,18 @@ void ainvfuzz(
                   pikp += xF[h];
                   // k's dam, k's dam's-group dp
                   aistart = pA[dp];
-                  for(i=aistart; i<pA[dp+1]; i++){
-                    if(iA[i] == d){
-                      xA[i] += alphai * pik * pikp;
-                      break;  			// break out of i for loop
+                  i = aistart;
+                  lb = pA[dp+1] - 1 - i;
+                  while(lb > 0){
+                    step = lb/2;
+                    it= i + step;
+                    if(iA[it] < d){
+                      i=++it;
+                      lb-=step+1;
                     }
-                  }  // end i for loop
-
+                    else lb = step;
+                  }
+                  if(iA[i] == d) xA[i] += alphai * pik * pikp;
                   break;  			// break out of h for loop
                 }
               }  // end h for loop
@@ -530,7 +617,7 @@ void ainvfuzz(
       aistart = pA[sj];
       // sire,sire
       xA[aistart] += alphai;
-      for(j=aistart; j<pA[sj+1]; j++){
+      for(j=aistart; j<pA[sj+1]; j++){  // NOTE: not using lower_bound algorithm
         // dam,sire
         if(iA[j] == dj){
           xA[j] += alphai;
@@ -545,7 +632,7 @@ void ainvfuzz(
       aistart = pA[dj];
       // dam,dam
       xA[aistart] += alphai;
-      for(j=aistart; j<pA[dj+1]; j++){
+      for(j=aistart; j<pA[dj+1]; j++){  // NOTE: not using lower_bound algorithm
         //sire,dam
         if(iA[j] == sj){
           xA[j] += alphai;
@@ -566,9 +653,8 @@ void ainvfuzz(
 
 
 
-
-
 ///////////////////////////////////////////////////////////////////////
+//	Used in makeA()
 ////   based on Hadfield's implementation of 
 ////   in MCMCglmm (nadiv < v2.13.4)
 
@@ -696,3 +782,4 @@ void acinv(
 }
 }
 
+
